"""Enumerations used in Datadoc."""

from __future__ import annotations

from enum import Enum

from dapla_metadata.datasets import enums
from dapla_metadata.datasets import model


class LanguageStringsEnum(Enum):
    """Enum class for storing LanguageStringType objects."""

    def __init__(
        self,
        language_strings: model.LanguageStringType,
    ) -> None:
        """Store the LanguageStringType object for displaying enum values in multiple languages.

        We don't particularly care what the value of the enum is,
        but when serialised it's convenient and readable to use the
        name of the enum, so we set the value to be the name.
        """
        self._value_ = self.name
        self.language_strings = language_strings

    def get_value_for_language(
        self,
        language: enums.SupportedLanguages,
    ) -> str | None:
        """Retrieve the string for the relevant language."""
        if self.language_strings.root is not None:
            for item in self.language_strings.root:
                if item.languageCode == language:
                    return item.languageText
        return None


class Assessment(LanguageStringsEnum):
    """Sensitivity of data."""

    SENSITIVE = model.LanguageStringType(
        [
            model.LanguageStringTypeItem(
                languageCode="en",
                languageText=model.Assessment.SENSITIVE.value,
            ),
            model.LanguageStringTypeItem(languageCode="nn", languageText="SENSITIV"),
            model.LanguageStringTypeItem(languageCode="nb", languageText="SENSITIV"),
        ],
    )
    PROTECTED = model.LanguageStringType(
        [
            model.LanguageStringTypeItem(
                languageCode="en",
                languageText=model.Assessment.PROTECTED.value,
            ),
            model.LanguageStringTypeItem(languageCode="nn", languageText="SKJERMET"),
            model.LanguageStringTypeItem(languageCode="nb", languageText="SKJERMET"),
        ],
    )
    OPEN = model.LanguageStringType(
        [
            model.LanguageStringTypeItem(
                languageCode="en",
                languageText=model.Assessment.OPEN.value,
            ),
            model.LanguageStringTypeItem(languageCode="nn", languageText="ÅPEN"),
            model.LanguageStringTypeItem(languageCode="nb", languageText="ÅPEN"),
        ],
    )


class DataSetStatus(LanguageStringsEnum):
    """Lifecycle status of a dataset."""

    DRAFT = model.LanguageStringType(
        [
            model.LanguageStringTypeItem(
                languageCode="en",
                languageText=model.DataSetStatus.DRAFT.value,
            ),
            model.LanguageStringTypeItem(languageCode="nn", languageText="UTKAST"),
            model.LanguageStringTypeItem(languageCode="nb", languageText="UTKAST"),
        ],
    )
    INTERNAL = model.LanguageStringType(
        [
            model.LanguageStringTypeItem(
                languageCode="en",
                languageText=model.DataSetStatus.INTERNAL.value,
            ),
            model.LanguageStringTypeItem(languageCode="nn", languageText="INTERN"),
            model.LanguageStringTypeItem(languageCode="nb", languageText="INTERN"),
        ],
    )
    EXTERNAL = model.LanguageStringType(
        [
            model.LanguageStringTypeItem(
                languageCode="en",
                languageText=model.DataSetStatus.EXTERNAL.value,
            ),
            model.LanguageStringTypeItem(languageCode="nn", languageText="EKSTERN"),
            model.LanguageStringTypeItem(languageCode="nb", languageText="EKSTERN"),
        ],
    )
    DEPRECATED = model.LanguageStringType(
        [
            model.LanguageStringTypeItem(
                languageCode="en",
                languageText=model.DataSetStatus.DEPRECATED.value,
            ),
            model.LanguageStringTypeItem(languageCode="nn", languageText="UTGÅTT"),
            model.LanguageStringTypeItem(languageCode="nb", languageText="UTGÅTT"),
        ],
    )


class DataSetState(LanguageStringsEnum):
    """Processing state of a dataset."""

    SOURCE_DATA = model.LanguageStringType(
        [
            model.LanguageStringTypeItem(
                languageCode="en",
                languageText=model.DataSetState.SOURCE_DATA.value,
            ),
            model.LanguageStringTypeItem(languageCode="nn", languageText="KILDEDATA"),
            model.LanguageStringTypeItem(languageCode="nb", languageText="KILDEDATA"),
        ],
    )
    INPUT_DATA = model.LanguageStringType(
        [
            model.LanguageStringTypeItem(
                languageCode="en",
                languageText=model.DataSetState.INPUT_DATA.value,
            ),
            model.LanguageStringTypeItem(languageCode="nn", languageText="INNDATA"),
            model.LanguageStringTypeItem(languageCode="nb", languageText="INNDATA"),
        ],
    )
    PROCESSED_DATA = model.LanguageStringType(
        [
            model.LanguageStringTypeItem(
                languageCode="en",
                languageText=model.DataSetState.PROCESSED_DATA.value,
            ),
            model.LanguageStringTypeItem(
                languageCode="nn",
                languageText="KLARGJORTE DATA",
            ),
            model.LanguageStringTypeItem(
                languageCode="nb",
                languageText="KLARGJORTE DATA",
            ),
        ],
    )
    STATISTICS = model.LanguageStringType(
        [
            model.LanguageStringTypeItem(
                languageCode="en",
                languageText=model.DataSetState.STATISTICS.value,
            ),
            model.LanguageStringTypeItem(languageCode="nn", languageText="STATISTIKK"),
            model.LanguageStringTypeItem(languageCode="nb", languageText="STATISTIKK"),
        ],
    )
    OUTPUT_DATA = model.LanguageStringType(
        [
            model.LanguageStringTypeItem(
                languageCode="en",
                languageText=model.DataSetState.OUTPUT_DATA.value,
            ),
            model.LanguageStringTypeItem(languageCode="nn", languageText="UTDATA"),
            model.LanguageStringTypeItem(languageCode="nb", languageText="UTDATA"),
        ],
    )


class TemporalityTypeType(LanguageStringsEnum):
    """Temporality of a dataset.

    More information about temporality type: https://statistics-norway.atlassian.net/l/c/HV12q90R
    """

    FIXED = model.LanguageStringType(
        [
            model.LanguageStringTypeItem(
                languageCode="en",
                languageText=model.TemporalityTypeType.FIXED.value,
            ),
            model.LanguageStringTypeItem(languageCode="nn", languageText="FAST"),
            model.LanguageStringTypeItem(languageCode="nb", languageText="FAST"),
        ],
    )
    STATUS = model.LanguageStringType(
        [
            model.LanguageStringTypeItem(
                languageCode="en",
                languageText=model.TemporalityTypeType.STATUS.value,
            ),
            model.LanguageStringTypeItem(languageCode="nn", languageText="TVERRSNITT"),
            model.LanguageStringTypeItem(languageCode="nb", languageText="TVERRSNITT"),
        ],
    )
    ACCUMULATED = model.LanguageStringType(
        [
            model.LanguageStringTypeItem(
                languageCode="en",
                languageText=model.TemporalityTypeType.ACCUMULATED.value,
            ),
            model.LanguageStringTypeItem(languageCode="nn", languageText="AKKUMULERT"),
            model.LanguageStringTypeItem(languageCode="nb", languageText="AKKUMULERT"),
        ],
    )
    EVENT = model.LanguageStringType(
        [
            model.LanguageStringTypeItem(
                languageCode="en",
                languageText=model.TemporalityTypeType.EVENT.value,
            ),
            model.LanguageStringTypeItem(languageCode="nn", languageText="HENDELSE"),
            model.LanguageStringTypeItem(languageCode="nb", languageText="HENDELSE"),
        ],
    )


class DataType(LanguageStringsEnum):
    """Simplified data types for metadata purposes."""

    STRING = model.LanguageStringType(
        [
            model.LanguageStringTypeItem(
                languageCode="en",
                languageText=model.DataType.STRING.value,
            ),
            model.LanguageStringTypeItem(languageCode="nn", languageText="TEKST"),
            model.LanguageStringTypeItem(languageCode="nb", languageText="TEKST"),
        ],
    )
    INTEGER = model.LanguageStringType(
        [
            model.LanguageStringTypeItem(
                languageCode="en",
                languageText=model.DataType.INTEGER.value,
            ),
            model.LanguageStringTypeItem(languageCode="nn", languageText="HELTALL"),
            model.LanguageStringTypeItem(languageCode="nb", languageText="HELTALL"),
        ],
    )
    FLOAT = model.LanguageStringType(
        [
            model.LanguageStringTypeItem(
                languageCode="en",
                languageText=model.DataType.FLOAT.value,
            ),
            model.LanguageStringTypeItem(languageCode="nn", languageText="DESIMALTALL"),
            model.LanguageStringTypeItem(languageCode="nb", languageText="DESIMALTALL"),
        ],
    )
    DATETIME = model.LanguageStringType(
        [
            model.LanguageStringTypeItem(
                languageCode="en",
                languageText=model.DataType.DATETIME.value,
            ),
            model.LanguageStringTypeItem(languageCode="nn", languageText="DATOTID"),
            model.LanguageStringTypeItem(languageCode="nb", languageText="DATOTID"),
        ],
    )
    BOOLEAN = model.LanguageStringType(
        [
            model.LanguageStringTypeItem(
                languageCode="en",
                languageText=model.DataType.BOOLEAN.value,
            ),
            model.LanguageStringTypeItem(languageCode="nn", languageText="BOOLSK"),
            model.LanguageStringTypeItem(languageCode="nb", languageText="BOOLSK"),
        ],
    )


class IsPersonalData(LanguageStringsEnum):
    """Is the variable instance personal data and if so, how is it encrypted."""

    NOT_PERSONAL_DATA = model.LanguageStringType(
        [
            model.LanguageStringTypeItem(
                languageCode="en",
                languageText=model.IsPersonalData.NOT_PERSONAL_DATA.value,
            ),
            model.LanguageStringTypeItem(
                languageCode="nn",
                languageText="IKKE PERSONOPPLYSNING",
            ),
            model.LanguageStringTypeItem(
                languageCode="nb",
                languageText="IKKE PERSONOPPLYSNING",
            ),
        ],
    )
    PSEUDONYMISED_ENCRYPTED_PERSONAL_DATA = model.LanguageStringType(
        [
            model.LanguageStringTypeItem(
                languageCode="en",
                languageText=model.IsPersonalData.PSEUDONYMISED_ENCRYPTED_PERSONAL_DATA.value,
            ),
            model.LanguageStringTypeItem(
                languageCode="nn",
                languageText="PSEUDONYMISERT/KRYPTERT PERSONOPPLYSNING",
            ),
            model.LanguageStringTypeItem(
                languageCode="nb",
                languageText="PSEUDONYMISERT/KRYPTERT PERSONOPPLYSNING",
            ),
        ],
    )
    NON_PSEUDONYMISED_ENCRYPTED_PERSONAL_DATA = model.LanguageStringType(
        [
            model.LanguageStringTypeItem(
                languageCode="en",
                languageText=model.IsPersonalData.NON_PSEUDONYMISED_ENCRYPTED_PERSONAL_DATA.value,
            ),
            model.LanguageStringTypeItem(
                languageCode="nn",
                languageText="IKKE PSEUDONYMISERT/KRYPTERT PERSONOPPLYSNING",
            ),
            model.LanguageStringTypeItem(
                languageCode="nb",
                languageText="IKKE PSEUDONYMISERT/KRYPTERT PERSONOPPLYSNING",
            ),
        ],
    )


class VariableRole(LanguageStringsEnum):
    """The role of a variable in a dataset."""

    IDENTIFIER = model.LanguageStringType(
        [
            model.LanguageStringTypeItem(
                languageCode="en",
                languageText=model.VariableRole.IDENTIFIER.value,
            ),
            model.LanguageStringTypeItem(
                languageCode="nn",
                languageText="IDENTIFIKATOR",
            ),
            model.LanguageStringTypeItem(
                languageCode="nb",
                languageText="IDENTIFIKATOR",
            ),
        ],
    )
    MEASURE = model.LanguageStringType(
        [
            model.LanguageStringTypeItem(
                languageCode="en",
                languageText=model.VariableRole.MEASURE.value,
            ),
            model.LanguageStringTypeItem(
                languageCode="nn",
                languageText="MÅLEVARIABEL",
            ),
            model.LanguageStringTypeItem(
                languageCode="nb",
                languageText="MÅLEVARIABEL",
            ),
        ],
    )
    START_TIME = model.LanguageStringType(
        [
            model.LanguageStringTypeItem(
                languageCode="en",
                languageText=model.VariableRole.START_TIME.value,
            ),
            model.LanguageStringTypeItem(languageCode="nn", languageText="STARTTID"),
            model.LanguageStringTypeItem(languageCode="nb", languageText="STARTTID"),
        ],
    )
    STOP_TIME = model.LanguageStringType(
        [
            model.LanguageStringTypeItem(
                languageCode="en",
                languageText=model.VariableRole.STOP_TIME.value,
            ),
            model.LanguageStringTypeItem(languageCode="nn", languageText="STOPPTID"),
            model.LanguageStringTypeItem(languageCode="nb", languageText="STOPPTID"),
        ],
    )
    ATTRIBUTE = model.LanguageStringType(
        [
            model.LanguageStringTypeItem(
                languageCode="en",
                languageText=model.VariableRole.ATTRIBUTE.value,
            ),
            model.LanguageStringTypeItem(languageCode="nn", languageText="ATTRIBUTT"),
            model.LanguageStringTypeItem(languageCode="nb", languageText="ATTRIBUTT"),
        ],
    )


class UseRestriction(LanguageStringsEnum):
    """Lifecycle status of a dataset."""

    DELETION_ANONYMIZATION = model.LanguageStringType(
        [
            model.LanguageStringTypeItem(
                languageCode="en",
                languageText=model.UseRestriction.DELETION_ANONYMIZATION.value,
            ),
            model.LanguageStringTypeItem(
                languageCode="nn",
                languageText="SLETTING/ANONYMISERING",
            ),
            model.LanguageStringTypeItem(
                languageCode="nb",
                languageText="SLETTING/ANONYMISERING",
            ),
        ],
    )
    PROCESS_LIMITATIONS = model.LanguageStringType(
        [
            model.LanguageStringTypeItem(
                languageCode="en",
                languageText=model.UseRestriction.PROCESS_LIMITATIONS.value,
            ),
            model.LanguageStringTypeItem(
                languageCode="nn",
                languageText="BEHANDLINGSBEGRENSNINGER",
            ),
            model.LanguageStringTypeItem(
                languageCode="nb",
                languageText="BEHANDLINGSBEGRENSNINGER",
            ),
        ],
    )
    SECONDARY_USE_RESTRICTIONS = model.LanguageStringType(
        [
            model.LanguageStringTypeItem(
                languageCode="en",
                languageText=model.UseRestriction.SECONDARY_USE_RESTRICTIONS.value,
            ),
            model.LanguageStringTypeItem(
                languageCode="nn",
                languageText="SEKUNDÆRBRUKSRESTRIKSJONER",
            ),
            model.LanguageStringTypeItem(
                languageCode="nb",
                languageText="SEKUNDÆRBRUKSRESTRIKSJONER",
            ),
        ],
    )



"""General utilities."""

from __future__ import annotations

import datetime
import importlib

METADATA_DOCUMENT_FILE_SUFFIX = "__DOC.json"


def running_in_notebook() -> bool:
    """Return True if running in Jupyter Notebook."""
    try:
        return bool(get_ipython().__class__.__name__ == "ZMQInteractiveShell")  # type: ignore [name-defined]
    except NameError:
        # The get_ipython method is globally available in ipython interpreters
        # as used in Jupyter. However it is not available in other python
        # interpreters and will throw a NameError. Therefore we're not running
        # in Jupyter.
        return False


def pick_random_port() -> int:
    """Pick a random free port number.

    The function will bind a socket to port 0, and a random free port from
    1024 to 65535 will be selected by the operating system.
    """
    import socket

    with socket.socket() as sock:
        sock.bind(("", 0))
        return int(sock.getsockname()[1])


def get_timestamp_now() -> datetime.datetime:
    """Return a timestamp for the current moment."""
    return datetime.datetime.now(tz=datetime.timezone.utc)


def get_app_version() -> str:
    """Get the version of the Datadoc package."""
    return importlib.metadata.distribution("ssb-datadoc").version



"""Functionality for displaying variables metadata."""

from __future__ import annotations

import functools
from enum import Enum

from dapla_metadata.datasets import enums

from datadoc import state
from datadoc.enums import DataType
from datadoc.enums import IsPersonalData
from datadoc.enums import TemporalityTypeType
from datadoc.enums import VariableRole
from datadoc.frontend.fields.display_base import VARIABLES_METADATA_DATE_INPUT
from datadoc.frontend.fields.display_base import VARIABLES_METADATA_MULTILANGUAGE_INPUT
from datadoc.frontend.fields.display_base import FieldTypes
from datadoc.frontend.fields.display_base import MetadataDropdownField
from datadoc.frontend.fields.display_base import MetadataInputField
from datadoc.frontend.fields.display_base import MetadataMultiLanguageField
from datadoc.frontend.fields.display_base import MetadataPeriodField
from datadoc.frontend.fields.display_base import get_data_source_options
from datadoc.frontend.fields.display_base import get_enum_options


def get_measurement_unit_options() -> list[dict[str, str]]:
    """Collect the unit type options."""
    dropdown_options = [
        {
            "title": measurement_unit.get_title(enums.SupportedLanguages.NORSK_BOKMÅL),
            "id": measurement_unit.code,
        }
        for measurement_unit in state.measurement_units.classifications
    ]
    dropdown_options.insert(0, {"title": "", "id": ""})
    return dropdown_options


class VariableIdentifiers(str, Enum):
    """As defined here: https://statistics-norway.atlassian.net/wiki/spaces/MPD/pages/3042869256/Variabelforekomst."""

    SHORT_NAME = "short_name"
    NAME = "name"
    DATA_TYPE = "data_type"
    VARIABLE_ROLE = "variable_role"
    DEFINITION_URI = "definition_uri"
    IS_PERSONAL_DATA = "is_personal_data"
    DATA_SOURCE = "data_source"
    POPULATION_DESCRIPTION = "population_description"
    COMMENT = "comment"
    TEMPORALITY_TYPE = "temporality_type"
    MEASUREMENT_UNIT = "measurement_unit"
    FORMAT = "format"
    CLASSIFICATION_URI = "classification_uri"
    INVALID_VALUE_DESCRIPTION = "invalid_value_description"
    IDENTIFIER = "id"
    CONTAINS_DATA_FROM = "contains_data_from"
    CONTAINS_DATA_UNTIL = "contains_data_until"
    DATA_ELEMENT_PATH = "data_element_path"
    MULTIPLICATION_FACTOR = "multiplication_factor"


DISPLAY_VARIABLES: dict[
    VariableIdentifiers,
    FieldTypes,
] = {
    VariableIdentifiers.NAME: MetadataMultiLanguageField(
        identifier=VariableIdentifiers.NAME.value,
        display_name="Navn",
        description="Variabelnavn som er forståelig for mennesker. Navnet kan arves fra  lenket VarDef-variabel eller endres her (ev. oppgis her i tilfeller der variabelen ikke skal lenkes til VarDef).",
        obligatory=True,
        id_type=VARIABLES_METADATA_MULTILANGUAGE_INPUT,
    ),
    VariableIdentifiers.DEFINITION_URI: MetadataInputField(
        identifier=VariableIdentifiers.DEFINITION_URI.value,
        display_name="Definition URI",
        description="Oppgi lenke (URI) til tilhørende variabel i VarDef.",
        obligatory=False,
    ),
    VariableIdentifiers.COMMENT: MetadataMultiLanguageField(
        identifier=VariableIdentifiers.COMMENT.value,
        display_name="Kommentar",
        description="Kommentaren har to funksjoner. Den skal brukes til å beskrive variabelforekomsten dersom denne ikke har lenke til VarDef (gjelder klargjorte data, statistikk og utdata), og den kan brukes til å gi ytterligere presiseringer av variabelforekomstens definisjon dersom variabelforekomsten er lenket til VarDef",
        id_type=VARIABLES_METADATA_MULTILANGUAGE_INPUT,
    ),
    VariableIdentifiers.IS_PERSONAL_DATA: MetadataDropdownField(
        identifier=VariableIdentifiers.IS_PERSONAL_DATA.value,
        display_name="Er personopplysning",
        description="Dersom variabelen er en personopplysning, skal det oppgis om den er pseudonymisert/kryptert eller ikke. Dersom den ikke er en personopplysning, lar en bare defaultsvaret «Ikke personopplysning» bli stående. All informasjon som entydig kan knyttes til en fysisk person (f.eks. fødselsnummer eller adresse) er personopplysninger. Næringsdata om enkeltpersonforetak (ENK) skal imidlertid ikke regnes som personopplysninger.",
        obligatory=True,
        options_getter=functools.partial(
            get_enum_options,
            IsPersonalData,
        ),
    ),
    VariableIdentifiers.POPULATION_DESCRIPTION: MetadataMultiLanguageField(
        identifier=VariableIdentifiers.POPULATION_DESCRIPTION.value,
        display_name="Populasjonen",
        description="Populasjonen settes på datasettnivå, men kan spesifiseres eller overskrives (hvis variabelen har en annen populasjon enn de fleste andre variablene i datasettet) her.",
        id_type=VARIABLES_METADATA_MULTILANGUAGE_INPUT,
    ),
    VariableIdentifiers.MEASUREMENT_UNIT: MetadataDropdownField(
        identifier=VariableIdentifiers.MEASUREMENT_UNIT.value,
        display_name="Måleenhet",
        description="Dersom variabelen er kvantitativ, skal den ha en måleenhet, f.eks. kilo eller kroner.",
        options_getter=get_measurement_unit_options,
    ),
    VariableIdentifiers.INVALID_VALUE_DESCRIPTION: MetadataMultiLanguageField(
        identifier=VariableIdentifiers.INVALID_VALUE_DESCRIPTION.value,
        display_name="Ugyldige verdier",
        description="Feltet brukes til å beskrive ugyldige verdier som inngår i variabelen - dersom spesialverdiene ikke er tilstrekkelige eller ikke kan benyttes.",
        id_type=VARIABLES_METADATA_MULTILANGUAGE_INPUT,
    ),
    VariableIdentifiers.MULTIPLICATION_FACTOR: MetadataInputField(
        identifier=VariableIdentifiers.MULTIPLICATION_FACTOR.value,
        display_name="Multiplikasjonsfaktor",
        description="Multiplikasjonsfaktoren er den numeriske verdien som multipliseres med måleenheten, f.eks. når en skal vise store tall i en tabell, eksempelvis 1000 kroner.",
        type="number",
    ),
    VariableIdentifiers.VARIABLE_ROLE: MetadataDropdownField(
        identifier=VariableIdentifiers.VARIABLE_ROLE.value,
        display_name="Variabelens rolle",
        description="Oppgi hvilken rolle variabelen har i datasettet. De ulike rollene er identifikator ( identifiserer de ulike enhetene, f.eks. fødselsnummer og organisasjonsnummer), målevariabel ( beskriver egenskaper, f.eks. sivilstand og omsetning), startdato (beskriver startdato for variabler som har et forløp, eller måletidspunkt for tverrsnittdata), stoppdato(beskriver stoppdato for variabler som har et forløp) og attributt (brukes i tifeller der SSB utvider datasettet med egen informasjon, f.eks. datakvalitet eller editering)",
        obligatory=True,
        options_getter=functools.partial(
            get_enum_options,
            VariableRole,
        ),
    ),
    VariableIdentifiers.CLASSIFICATION_URI: MetadataInputField(
        identifier=VariableIdentifiers.CLASSIFICATION_URI.value,
        display_name="Kodeverkets URI",
        description="Lenke (URI) til gyldige kodeverk (klassifikasjon eller kodeliste) i KLASS eller Klass-uttrekk. Variabelforekomster skal generelt knyttes til tilhørende kodeverk via relevant variabeldefinisjon i Vardef. Unntaksvis kan den imidlertid knyttes direkte til Klass via dette feltet (i tilfeller der det ikke er hensiktsmessig å definere variabelen i Vardef).",
    ),
    VariableIdentifiers.DATA_SOURCE: MetadataDropdownField(
        identifier=VariableIdentifiers.DATA_SOURCE.value,
        display_name="Datakilde",
        description="Oppgi datakilden til variabelen (på etat-/organisasjonsnivå) dersom denne ikke allerede er satt på  datasettnivå. Brukes hovedsakelig når variabler i et datasett har ulike datakilder.",
        options_getter=get_data_source_options,
    ),
    VariableIdentifiers.TEMPORALITY_TYPE: MetadataDropdownField(
        identifier=VariableIdentifiers.TEMPORALITY_TYPE.value,
        display_name="Temporalitetstype",
        description="Temporalitetstypen settes vanligvis på datasettnivå, men dersom datasettet består av variabler med ulike temporalitetstyper, kan den settes på variabelnivå. Temporalitet sier noe om tidsdimensjonen i datasettet. Fast er data med verdi som ikke endres over tid (f.eks. fødselsdato), tverrsnitt er data som er målt på et gitt tidspunkt, akkumulert er data som  er samlet over en viss tidsperiode (f.eks. inntekt gjennom et år) og  hendelse/forløp registrerer tidspunkt og tidsperiode for ulike hendelser /tilstander, f.eks. (skifte av) bosted.",
        options_getter=functools.partial(
            get_enum_options,
            TemporalityTypeType,
        ),
    ),
    VariableIdentifiers.FORMAT: MetadataInputField(
        identifier=VariableIdentifiers.FORMAT.value,
        display_name="Format",
        description="Verdienes format (fysisk format eller regulært uttrykk) i maskinlesbar form ifm validering, f.eks.  ISO 8601 som datoformat. Dette feltet kan benyttes som en ytterligere presisering av datatype i tilfellene der det er relevant.",
    ),
    VariableIdentifiers.CONTAINS_DATA_FROM: MetadataPeriodField(
        identifier=VariableIdentifiers.CONTAINS_DATA_FROM.value,
        display_name="Inneholder data f.o.m.",
        description="Variabelen inneholder data fra og med denne datoen.",
        id_type=VARIABLES_METADATA_DATE_INPUT,
    ),
    VariableIdentifiers.CONTAINS_DATA_UNTIL: MetadataPeriodField(
        identifier=VariableIdentifiers.CONTAINS_DATA_UNTIL.value,
        display_name="Inneholder data t.o.m.",
        description="Variabelen inneholder data til og med denne datoen.",
        id_type=VARIABLES_METADATA_DATE_INPUT,
    ),
    VariableIdentifiers.SHORT_NAME: MetadataInputField(
        identifier=VariableIdentifiers.SHORT_NAME.value,
        display_name="Kortnavn",
        description="Fysisk navn på variabelen i datasettet. Bør tilsvare anbefalt kortnavn.",
        obligatory=True,
        editable=False,
    ),
    VariableIdentifiers.DATA_TYPE: MetadataDropdownField(
        identifier=VariableIdentifiers.DATA_TYPE.value,
        display_name="Datatype",
        description="Velg en av følgende datatyper: tekst, heltall, desimaltall, datotid eller boolsk. Dersom variabelen er knyttet til et kodeverk i Klass, velges datatype tekst.",
        obligatory=True,
        options_getter=functools.partial(
            get_enum_options,
            DataType,
        ),
    ),
    VariableIdentifiers.DATA_ELEMENT_PATH: MetadataInputField(
        identifier=VariableIdentifiers.DATA_ELEMENT_PATH.value,
        display_name="Dataelementsti",
        description="For hierarkiske datasett (JSON) må sti til dataelementet oppgis i tillegg til kortnavn (shortName)",
    ),
    VariableIdentifiers.IDENTIFIER: MetadataInputField(
        identifier=VariableIdentifiers.IDENTIFIER.value,
        display_name="ID",
        description="Unik SSB identifikator for variabelforekomsten i datasettet",
        editable=False,
    ),
}

MULTIPLE_LANGUAGE_VARIABLES_METADATA = [
    m.identifier
    for m in DISPLAY_VARIABLES.values()
    if isinstance(m, MetadataMultiLanguageField)
]

VARIABLES_METADATA_LEFT = [
    m
    for m in DISPLAY_VARIABLES.values()
    if (isinstance(m, MetadataMultiLanguageField) and m.editable)
]

VARIABLES_METADATA_RIGHT = [
    m
    for m in DISPLAY_VARIABLES.values()
    if not isinstance(m, MetadataMultiLanguageField) and m.editable
]

OBLIGATORY_VARIABLES_METADATA = [
    m for m in DISPLAY_VARIABLES.values() if m.obligatory and m.editable
]

OPTIONAL_VARIABLES_METADATA = [
    m for m in DISPLAY_VARIABLES.values() if not m.obligatory and m.editable
]

OBLIGATORY_VARIABLES_METADATA_IDENTIFIERS_AND_DISPLAY_NAME: list[tuple] = [
    (m.identifier, m.display_name)
    for m in DISPLAY_VARIABLES.values()
    if m.obligatory and m.editable
]

NON_EDITABLE_VARIABLES_METADATA = [
    m for m in DISPLAY_VARIABLES.values() if not m.editable
]



"""All decorated callback functions should be defined here.

Implementations of the callback functionality should be in other functions (in other files), to enable unit testing.
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from dash import MATCH
from dash import Dash
from dash import Input
from dash import Output
from dash import State
from dash import ctx
from dash import html
from dash import no_update

from datadoc import state
from datadoc.frontend.callbacks.dataset import accept_dataset_metadata_date_input
from datadoc.frontend.callbacks.dataset import accept_dataset_metadata_input
from datadoc.frontend.callbacks.dataset import open_dataset_handling
from datadoc.frontend.callbacks.utils import render_tabs
from datadoc.frontend.callbacks.utils import save_metadata_and_generate_alerts
from datadoc.frontend.callbacks.variables import accept_variable_metadata_date_input
from datadoc.frontend.callbacks.variables import accept_variable_metadata_input
from datadoc.frontend.callbacks.variables import populate_variables_workspace
from datadoc.frontend.components.builders import build_dataset_edit_section
from datadoc.frontend.components.builders import build_dataset_machine_section
from datadoc.frontend.components.identifiers import ACCORDION_WRAPPER_ID
from datadoc.frontend.components.identifiers import SECTION_WRAPPER_ID
from datadoc.frontend.components.identifiers import VARIABLES_INFORMATION_ID
from datadoc.frontend.fields.display_base import DATASET_METADATA_DATE_INPUT
from datadoc.frontend.fields.display_base import DATASET_METADATA_INPUT
from datadoc.frontend.fields.display_base import DATASET_METADATA_MULTILANGUAGE_INPUT
from datadoc.frontend.fields.display_base import VARIABLES_METADATA_DATE_INPUT
from datadoc.frontend.fields.display_base import VARIABLES_METADATA_INPUT
from datadoc.frontend.fields.display_base import VARIABLES_METADATA_MULTILANGUAGE_INPUT
from datadoc.frontend.fields.display_dataset import EDITABLE_DATASET_METADATA_LEFT
from datadoc.frontend.fields.display_dataset import EDITABLE_DATASET_METADATA_RIGHT
from datadoc.frontend.fields.display_dataset import NON_EDITABLE_DATASET_METADATA
from datadoc.frontend.fields.display_dataset import DatasetIdentifiers
from datadoc.frontend.fields.display_variables import VariableIdentifiers

if TYPE_CHECKING:
    import dash_bootstrap_components as dbc

    from datadoc.frontend.callbacks.utils import MetadataInputTypes

logger = logging.getLogger(__name__)


def register_callbacks(app: Dash) -> None:
    """Define and register callbacks."""

    @app.callback(
        Output("alerts-section", "children", allow_duplicate=True),
        Input("save-button", "n_clicks"),
        State("alerts-section", "children"),
        prevent_initial_call=True,
    )
    def callback_save_metadata_file(
        n_clicks: int,
        alerts: list,  # argument required by Dash  # noqa: ARG001
    ) -> list:
        """Save the metadata document to disk and check obligatory metadata.

        Returns:
            List of alerts. Obligatory metadata alert warning if there is obligatory metadata missing.
            And success alert if metadata is saved correctly.
            If none return no_update.
        """
        if n_clicks and n_clicks > 0:
            return save_metadata_and_generate_alerts(state.metadata)

        return no_update

    @app.callback(
        Output(
            {"type": DATASET_METADATA_INPUT, "id": MATCH},
            "error",
        ),
        Output(
            {"type": DATASET_METADATA_INPUT, "id": MATCH},
            "errorMessage",
        ),
        Input(
            {"type": DATASET_METADATA_INPUT, "id": MATCH},
            "value",
        ),
        prevent_initial_call=True,
    )
    def callback_accept_dataset_metadata_input(
        value: MetadataInputTypes,  # noqa: ARG001 argument required by Dash
    ) -> tuple[bool, str]:
        """Save updated dataset metadata values.

        Will display an alert if validation fails.
        """
        return accept_dataset_metadata_input(
            ctx.triggered[0]["value"],
            ctx.triggered_id["id"],
        )

    @app.callback(
        Output(
            {
                "type": DATASET_METADATA_MULTILANGUAGE_INPUT,
                "id": MATCH,
                "language": MATCH,
            },
            "error",
        ),
        Output(
            {
                "type": DATASET_METADATA_MULTILANGUAGE_INPUT,
                "id": MATCH,
                "language": MATCH,
            },
            "errorMessage",
        ),
        Input(
            {
                "type": DATASET_METADATA_MULTILANGUAGE_INPUT,
                "id": MATCH,
                "language": MATCH,
            },
            "value",
        ),
        prevent_initial_call=True,
    )
    def callback_accept_dataset_metadata_multilanguage_input(
        value: MetadataInputTypes,  # noqa: ARG001 argument required by Dash
    ) -> tuple[bool, str]:
        """Save updated dataset metadata values.

        Will display an alert if validation fails.
        """
        # Get the ID of the input that changed. This MUST match the attribute name defined in DataDocDataSet
        return accept_dataset_metadata_input(
            ctx.triggered[0]["value"],
            ctx.triggered_id["id"],
            ctx.triggered_id["language"],
        )

    @app.callback(
        Output("alerts-section", "children", allow_duplicate=True),
        Output("dataset-opened-counter", "data"),  # Used to force reload of metadata
        Input("open-button", "n_clicks"),
        State("dataset-path-input", "value"),
        State("dataset-opened-counter", "data"),
        prevent_initial_call=True,
    )
    def callback_open_dataset(
        n_clicks: int,
        dataset_path: str,
        dataset_opened_counter: int,
    ) -> tuple[dbc.Alert, int]:
        """Open a dataset.

        Shows an alert on success or failure.

        To trigger reload of data in the UI, we update the
        language dropdown. This is a hack and could be replaced
        by a more formal mechanism.
        """
        return open_dataset_handling(n_clicks, dataset_path, dataset_opened_counter)

    @app.callback(
        Output("display-tab", "children"),
        Input("tabs", "value"),
    )
    def callback_render_tabs(tab: html.Article) -> html.Article | None:
        """Return correct tab content."""
        return render_tabs(tab)

    @app.callback(
        Output(VARIABLES_INFORMATION_ID, "children"),
        Input("dataset-opened-counter", "data"),
    )
    def callback_populate_variables_info_section(
        dataset_opened_counter: int,  # noqa: ARG001 Dash requires arguments for all Inputs
    ) -> str:
        if state.metadata.variables and len(state.metadata.variables) > 0:
            return f"Datasettet inneholder {len(state.metadata.variables)} variabler."

        return "Åpne et datasett for å liste variablene."

    @app.callback(
        Output(ACCORDION_WRAPPER_ID, "children"),
        Input("dataset-opened-counter", "data"),
        Input("search-variables", "value"),
    )
    def callback_populate_variables_workspace(
        dataset_opened_counter: int,  # Dash requires arguments for all Inputs
        search_query: str,
    ) -> list:
        """Create variable workspace with accordions for variables.

        Allows for filtering which variables are displayed via the search box.
        """
        logger.debug("Populating variables workspace. Search query: %s", search_query)
        return populate_variables_workspace(
            state.metadata.variables,
            search_query,
            dataset_opened_counter,
        )

    @app.callback(
        Output(SECTION_WRAPPER_ID, "children"),
        Input("dataset-opened-counter", "data"),
    )
    def callback_populate_dataset_workspace(
        dataset_opened_counter: int,  # Dash requires arguments for all Inputs
    ) -> list:
        """Create dataset workspace with sections."""
        logger.debug("Populating dataset workspace")
        return [
            build_dataset_edit_section(
                [
                    EDITABLE_DATASET_METADATA_LEFT,
                    EDITABLE_DATASET_METADATA_RIGHT,
                ],
                state.metadata.dataset,
                {
                    "type": "dataset-edit-section",
                    "id": f"obligatory-{dataset_opened_counter}",
                },
            ),
            build_dataset_machine_section(
                "Maskingenerert",
                NON_EDITABLE_DATASET_METADATA,
                state.metadata.dataset,
                {
                    "type": "dataset-machine-section",
                    "id": f"machine-{dataset_opened_counter}",
                },
            ),
        ]

    @app.callback(
        Output(
            {
                "type": VARIABLES_METADATA_INPUT,
                "variable_short_name": MATCH,
                "id": MATCH,
            },
            "error",
        ),
        Output(
            {
                "type": VARIABLES_METADATA_INPUT,
                "variable_short_name": MATCH,
                "id": MATCH,
            },
            "errorMessage",
        ),
        Input(
            {
                "type": VARIABLES_METADATA_INPUT,
                "variable_short_name": MATCH,
                "id": MATCH,
            },
            "value",
        ),
        prevent_initial_call=True,
    )
    def callback_accept_variable_metadata_input(
        value: MetadataInputTypes,  # noqa: ARG001 argument required by Dash
    ) -> dbc.Alert:
        """Save updated variable metadata values."""
        message = accept_variable_metadata_input(
            ctx.triggered[0]["value"],
            ctx.triggered_id["variable_short_name"],
            ctx.triggered_id["id"],
        )
        if not message:
            # No error to display.
            return False, ""

        return True, message

    @app.callback(
        Output(
            {
                "type": VARIABLES_METADATA_MULTILANGUAGE_INPUT,
                "variable_short_name": MATCH,
                "id": MATCH,
                "language": MATCH,
            },
            "error",
        ),
        Output(
            {
                "type": VARIABLES_METADATA_MULTILANGUAGE_INPUT,
                "variable_short_name": MATCH,
                "id": MATCH,
                "language": MATCH,
            },
            "errorMessage",
        ),
        Input(
            {
                "type": VARIABLES_METADATA_MULTILANGUAGE_INPUT,
                "variable_short_name": MATCH,
                "id": MATCH,
                "language": MATCH,
            },
            "value",
        ),
        prevent_initial_call=True,
    )
    def callback_accept_variable_metadata_multilanguage_input(
        value: MetadataInputTypes,  # noqa: ARG001 argument required by Dash
    ) -> dbc.Alert:
        """Save updated variable metadata values."""
        message = accept_variable_metadata_input(
            ctx.triggered[0]["value"],
            ctx.triggered_id["variable_short_name"],
            ctx.triggered_id["id"],
            ctx.triggered_id["language"],
        )
        if not message:
            # No error to display.
            return False, ""

        return True, message

    @app.callback(
        Output(
            {
                "type": VARIABLES_METADATA_DATE_INPUT,
                "variable_short_name": MATCH,
                "id": VariableIdentifiers.CONTAINS_DATA_FROM.value,
            },
            "error",
        ),
        Output(
            {
                "type": VARIABLES_METADATA_DATE_INPUT,
                "variable_short_name": MATCH,
                "id": VariableIdentifiers.CONTAINS_DATA_FROM.value,
            },
            "errorMessage",
        ),
        Output(
            {
                "type": VARIABLES_METADATA_DATE_INPUT,
                "variable_short_name": MATCH,
                "id": VariableIdentifiers.CONTAINS_DATA_UNTIL.value,
            },
            "error",
        ),
        Output(
            {
                "type": VARIABLES_METADATA_DATE_INPUT,
                "variable_short_name": MATCH,
                "id": VariableIdentifiers.CONTAINS_DATA_UNTIL.value,
            },
            "errorMessage",
        ),
        Input(
            {
                "type": VARIABLES_METADATA_DATE_INPUT,
                "variable_short_name": MATCH,
                "id": VariableIdentifiers.CONTAINS_DATA_FROM.value,
            },
            "value",
        ),
        Input(
            {
                "type": VARIABLES_METADATA_DATE_INPUT,
                "variable_short_name": MATCH,
                "id": VariableIdentifiers.CONTAINS_DATA_UNTIL.value,
            },
            "value",
        ),
        prevent_initial_call=True,
    )
    def callback_accept_variable_metadata_date_input(
        contains_data_from: str,
        contains_data_until: str,
    ) -> dbc.Alert:
        """Special case handling for date fields which have a relationship to one another."""
        return accept_variable_metadata_date_input(
            VariableIdentifiers(ctx.triggered_id["id"]),
            ctx.triggered_id["variable_short_name"],
            contains_data_from,
            contains_data_until,
        )

    @app.callback(
        Output(
            {
                "type": DATASET_METADATA_DATE_INPUT,
                "id": DatasetIdentifiers.CONTAINS_DATA_FROM.value,
            },
            "error",
        ),
        Output(
            {
                "type": DATASET_METADATA_DATE_INPUT,
                "id": DatasetIdentifiers.CONTAINS_DATA_FROM.value,
            },
            "errorMessage",
        ),
        Output(
            {
                "type": DATASET_METADATA_DATE_INPUT,
                "id": DatasetIdentifiers.CONTAINS_DATA_UNTIL.value,
            },
            "error",
        ),
        Output(
            {
                "type": DATASET_METADATA_DATE_INPUT,
                "id": DatasetIdentifiers.CONTAINS_DATA_UNTIL.value,
            },
            "errorMessage",
        ),
        Input(
            {
                "type": DATASET_METADATA_DATE_INPUT,
                "id": DatasetIdentifiers.CONTAINS_DATA_FROM.value,
            },
            "value",
        ),
        Input(
            {
                "type": DATASET_METADATA_DATE_INPUT,
                "id": DatasetIdentifiers.CONTAINS_DATA_UNTIL.value,
            },
            "value",
        ),
        prevent_initial_call=True,
    )
    def callback_accept_dataset_metadata_date_input(
        contains_data_from: str,
        contains_data_until: str,
    ) -> dbc.Alert:
        """Special case handling for date fields which have a relationship to one another."""
        return accept_dataset_metadata_date_input(
            DatasetIdentifiers(ctx.triggered_id["id"]),
            contains_data_from,
            contains_data_until,
        )



from __future__ import annotations

import datetime as dt
import json
import logging
from typing import Any


class DatadocJSONFormatter(logging.Formatter):
    """Class for formatting json for log files."""

    def __init__(
        self,
        *,
        fmt_keys: dict[str, str] | None = None,
    ) -> None:
        """Initializer for the json formatter."""
        super().__init__()
        self.fmt_keys = fmt_keys if fmt_keys is not None else {}

    def format(self, record: logging.LogRecord) -> str:
        """Method that creates the json structure from a message created by the _prepare_log_dict method."""
        message = self._prepare_log_dict(record)
        return json.dumps(message, default=str)

    def _prepare_log_dict(self, record: logging.LogRecord) -> dict[str, str | Any]:
        always_fields = {
            "message": record.getMessage(),
            "timestamp": dt.datetime.fromtimestamp(
                record.created,
                tz=dt.timezone.utc,
            ).isoformat(),
        }
        if record.exc_info is not None:
            always_fields["exc_info"] = self.formatException(record.exc_info)

        if record.stack_info is not None:
            always_fields["stack_info"] = self.formatStack(record.stack_info)

        message = {
            key: (
                msg_val
                if (msg_val := always_fields.pop(val, None)) is not None
                else getattr(record, val)
            )
            for key, val in self.fmt_keys.items()
        }
        message.update(always_fields)

        return message



"""Functions which aren't directly called from a decorated callback."""

from __future__ import annotations

import datetime
import json
import logging
import re
import warnings
from typing import TYPE_CHECKING
from typing import TypeAlias

import arrow
import ssb_dash_components as ssb
from dapla_metadata.datasets import Datadoc
from dapla_metadata.datasets import ObligatoryDatasetWarning
from dapla_metadata.datasets import ObligatoryVariableWarning
from dapla_metadata.datasets import model
from dash import html

from datadoc import config
from datadoc import state
from datadoc.constants import CHECK_OBLIGATORY_METADATA_DATASET_MESSAGE
from datadoc.constants import CHECK_OBLIGATORY_METADATA_VARIABLES_MESSAGE
from datadoc.constants import MISSING_METADATA_WARNING
from datadoc.frontend.components.builders import AlertTypes
from datadoc.frontend.components.builders import build_ssb_alert
from datadoc.frontend.components.identifiers import ACCORDION_WRAPPER_ID
from datadoc.frontend.components.identifiers import SECTION_WRAPPER_ID
from datadoc.frontend.components.identifiers import VARIABLES_INFORMATION_ID
from datadoc.frontend.fields.display_dataset import (
    OBLIGATORY_DATASET_METADATA_IDENTIFIERS_AND_DISPLAY_NAME,
)
from datadoc.frontend.fields.display_variables import (
    OBLIGATORY_VARIABLES_METADATA_IDENTIFIERS_AND_DISPLAY_NAME,
)

if TYPE_CHECKING:
    import pathlib

    import dash_bootstrap_components as dbc
    import pydantic
    from cloudpathlib import CloudPath


logger = logging.getLogger(__name__)


MetadataInputTypes: TypeAlias = (
    str | list[str] | int | float | bool | datetime.date | None
)


def _check_if_language_string_item_exists(
    language_strings: model.LanguageStringType,
    language_code: str,
) -> bool:
    if language_strings.root is None:
        return False
    return any(i.languageCode == language_code for i in language_strings.root)


def _update_language_string_item(
    language_strings: model.LanguageStringType,
    language_code: str,
    new_value: str,
) -> None:
    if language_strings.root is not None:
        for i in language_strings.root:
            if i.languageCode == language_code:
                i.languageText = new_value


def _add_language_string_item(
    language_strings: model.LanguageStringType,
    language_code: str,
    language_text: str,
) -> None:
    if language_strings.root is not None:
        language_strings.root.append(
            model.LanguageStringTypeItem(
                languageCode=language_code,
                languageText=language_text,
            ),
        )


def find_existing_language_string(
    metadata_model_object: pydantic.BaseModel,
    value: str,
    metadata_identifier: str,
    language: str,
) -> model.LanguageStringType | None:
    """Get or create a LanguageStrings object and return it."""
    language_strings = getattr(metadata_model_object, metadata_identifier)

    if language_strings is not None:
        if _check_if_language_string_item_exists(
            language_strings,
            language,
        ):
            _update_language_string_item(
                language_strings,
                language,
                value,
            )
        elif value != "":
            _add_language_string_item(
                language_strings,
                language,
                value,
            )
        else:
            return None
    elif value != "":
        language_strings = model.LanguageStringType(
            root=[
                model.LanguageStringTypeItem(
                    languageCode=language,
                    languageText=value,
                ),
            ],
        )
    else:
        # Don't create an item if the value is empty
        return None
    return language_strings


def get_dataset_path() -> pathlib.Path | CloudPath | str:
    """Extract the path to the dataset from the potential sources."""
    if state.metadata.dataset_path is not None:
        return state.metadata.dataset_path
    path_from_env = config.get_datadoc_dataset_path()
    if path_from_env:
        logger.info(
            "Dataset path from env var: '%s'",
            path_from_env,
        )
    if path_from_env is None:
        path_from_env = ""
    return path_from_env


VALIDATION_ERROR = "Validation error: "
DATE_VALIDATION_MESSAGE = f"{VALIDATION_ERROR}contains_data_from must be the same or earlier date than contains_data_until"


def parse_and_validate_dates(
    start_date: str | datetime.datetime | None,
    end_date: str | datetime.datetime | None,
) -> tuple[datetime.datetime | None, datetime.datetime | None]:
    """Parse and validate the given dates.

    Validate that:
        - The dates are in YYYY-MM-DD format
        - The start date is earlier or identical to the end date.

    Examples:
    >>> parse_and_validate_dates("2021-01-01", "2021-01-01")
    (datetime.datetime(2021, 1, 1, 0, 0, tzinfo=datetime.timezone.utc), datetime.datetime(2021, 1, 1, 0, 0, tzinfo=datetime.timezone.utc))

    >>> parse_and_validate_dates("1990-01-01", "2050-01-01")
    (datetime.datetime(1990, 1, 1, 0, 0, tzinfo=datetime.timezone.utc), datetime.datetime(2050, 1, 1, 0, 0, tzinfo=datetime.timezone.utc))

    >>> parse_and_validate_dates(None, None)
    (None, None)

    >>> parse_and_validate_dates("1st January 2021", "1st January 2021")
    Traceback (most recent call last):
    ...
    ValueError: Validation error: Expected an ISO 8601-like string, but was given '1st January 2021'. Try passing in a format string to resolve this.

    >>> parse_and_validate_dates(datetime.datetime(2050, 1, 1, 0, 0, tzinfo=datetime.timezone.utc), "1990-01-01")
    Traceback (most recent call last):
    ...
    ValueError: Validation error: contains_data_from must be the same or earlier date than contains_data_until

    >>> parse_and_validate_dates("2050-01-01", "1990-01-01")
    Traceback (most recent call last):
    ...
    ValueError: Validation error: contains_data_from must be the same or earlier date than contains_data_until
    """
    parsed_start = None
    parsed_end = None
    try:
        if start_date and start_date != "None":
            parsed_start = arrow.get(start_date)
        if end_date and end_date != "None":
            parsed_end = arrow.get(end_date)
    except arrow.parser.ParserError as e:
        raise ValueError(VALIDATION_ERROR + str(e)) from e

    if parsed_start and parsed_end and (parsed_start > parsed_end):
        raise ValueError(DATE_VALIDATION_MESSAGE)

    start_output = (
        parsed_start.astimezone(tz=datetime.timezone.utc) if parsed_start else None
    )
    end_output = parsed_end.astimezone(tz=datetime.timezone.utc) if parsed_end else None

    return start_output, end_output


def render_tabs(tab: str) -> html.Article | None:
    """Render tab content."""
    if tab == "dataset":
        return html.Article(
            [
                html.Article(
                    id=SECTION_WRAPPER_ID,
                    className="workspace-content",
                ),
            ],
            className="workspace-page-wrapper",
        )
    if tab == "variables":
        return html.Article(
            [
                html.Header(
                    [
                        ssb.Paragraph(
                            id=VARIABLES_INFORMATION_ID,
                            className="workspace-info-paragraph",
                        ),
                        ssb.Input(
                            label="Filtrer",
                            searchField=True,
                            disabled=False,
                            placeholder="Variabel kortnavn...",
                            id="search-variables",
                            n_submit=0,
                            value="",
                        ),
                    ],
                    className="workspace-header",
                ),
                html.Article(
                    id=ACCORDION_WRAPPER_ID,
                    className="workspace-content",
                ),
            ],
            className="workspace-page-wrapper",
        )

    return None


def _has_exact_word(word: str, text: str) -> bool:
    """Return True if excact word matches text."""
    return bool(re.search(rf"\b{word}\b", text))


def dataset_control(error_message: str | None) -> dbc.Alert | None:
    """Check obligatory metadata values for dataset.

    Args:
        error_message(str): A message generated by ObligatoryDatasetWarning containing names of fields missing value.
    """
    missing_metadata = [
        f[1]
        for f in OBLIGATORY_DATASET_METADATA_IDENTIFIERS_AND_DISPLAY_NAME
        if (error_message and _has_exact_word(f[0], error_message))
    ]
    if not missing_metadata:
        return None
    return build_ssb_alert(
        AlertTypes.WARNING,
        MISSING_METADATA_WARNING,
        CHECK_OBLIGATORY_METADATA_DATASET_MESSAGE,
        None,
        missing_metadata,
    )


def _parse_error_message(message: str) -> list | None:
    """Parse a string containing an error message into Python objects.

    This function processes an error message by removing predefined text and
    attempting to parse the remaining string into a list of Python objects.

    Args:
        message (str): The error message string to be parsed.

    Returns:
        A list of parsed objects if the string can be successfully parsed.
        returns None if the string is empty after processing and a empty list
        if the string cannot be parsed into JSON.
    """
    parsed_string = message.replace("Obligatory metadata is missing: ", "").strip()
    parsed_string = parsed_string.replace("'", '"')
    if not parsed_string:
        return None
    try:
        # Attempt to parse the JSON string
        return json.loads(parsed_string)
    except json.JSONDecodeError:
        logger.exception("Error parsing JSON {e}")
        return []


def _get_dict_by_key(
    metadata_list: list[dict[str, list[str]]],
    key: str,
) -> dict[str, list[str]] | None:
    """Return the first dictionary containing the specified key.

    This function searches through a list of dictionaries and returns the first
    dictionary that contains the specified key.

    Args:
        metadata_list: A list of dictionaries to search.
        key: The key to search for in each dictionary.

    Returns:
        The first dictionary containing the specified key,
        or None if no such dictionary is found.

    """
    return next((item for item in metadata_list if key in item), None)


def variables_control(error_message: str | None, variables: list) -> dbc.Alert | None:
    """Check obligatory metadata for variables and return an alert if any metadata is missing.

    This function parses an error message to identify missing obligatory metadata
    fields for variables. If missing metadata is found, it generates an alert.

    Args:
        error_message: A message generated by ObligatoryVariableWarning
            containing the variable short name and a list of field names with missing values.
        variables: list of datadoc variables

    Returns:
        An alert object if there are missing metadata fields, otherwise None.
    """
    missing_metadata: list = []
    error_message_parsed = _parse_error_message(str(error_message))
    # for variable in state.metadata.variables:
    for variable in variables:
        if error_message_parsed:
            fields_by_variable = _get_dict_by_key(
                error_message_parsed,
                variable.short_name,
            )
            if fields_by_variable is not None:
                missing_metadata_field = [
                    f[1]
                    for f in OBLIGATORY_VARIABLES_METADATA_IDENTIFIERS_AND_DISPLAY_NAME
                    if error_message and f[0] in fields_by_variable[variable.short_name]
                ]
                missing_metadata_fields_to_string = ", ".join(missing_metadata_field)
                missing_metadata.append(
                    f"{variable.short_name}: {missing_metadata_fields_to_string}",
                )
    if not missing_metadata:
        return None
    return build_ssb_alert(
        AlertTypes.WARNING,
        MISSING_METADATA_WARNING,
        CHECK_OBLIGATORY_METADATA_VARIABLES_MESSAGE,
        None,
        missing_metadata,
    )


def save_metadata_and_generate_alerts(metadata: Datadoc) -> list:
    """Save the metadata document to disk and check obligatory metadata.

    Returns:
        List of alerts including obligatory metadata warnings if missing,
        and success alert if metadata is saved correctly.
    """
    missing_obligatory_dataset = ""
    missing_obligatory_variables = ""

    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        metadata.write_metadata_document()
        success_alert = build_ssb_alert(
            AlertTypes.SUCCESS,
            "Lagret metadata",
        )

        for warning in w:
            if issubclass(warning.category, ObligatoryDatasetWarning):
                missing_obligatory_dataset = str(warning.message)
            elif issubclass(warning.category, ObligatoryVariableWarning):
                missing_obligatory_variables = str(warning.message)
            else:
                logger.warning(
                    "An unexpected warning was caught: %s",
                    warning.message,
                )

    return [
        success_alert,
        dataset_control(missing_obligatory_dataset),
        variables_control(missing_obligatory_variables, metadata.variables),
    ]



"""Repository for constant values in Datadoc frontend module."""

INVALID_VALUE = "Ugyldig verdi angitt!"
INVALID_DATE_ORDER = "Verdien for {contains_data_from_display_name} må være en lik eller tidligere dato som {contains_data_until_display_name}"
